import logging
from .notification import Notification


class LogNotification(Notification):
    """Logs via log_cfg, and also to a notification file with the format'[date time][self.name] msg'"""

    def __init__(self, name, log_file):
        super().__init__(name)
        handler = logging.FileHandler(filename=log_file, mode='a')
        handler.setFormatter(
            logging.Formatter(
                fmt='[%(asctime)s][' + self.name + '] %(message)s',
                datefmt='%Y-%b-%d %H:%M:%S'
            )
        )
        handler.setLevel(logging.INFO)
        self._logger.addHandler(handler)

    def notify(self, msg):
        self.info(msg)



from unittest.mock import patch

import os

from egcg_core.archive_management import archive_states, release_file_from_lustre, ArchivingError, \
    register_for_archiving, recall_from_tape, archive_directory
from tests import TestEGCG


class TestArchiveManagement(TestEGCG):

    def test_archive_states(self):

        with patch('egcg_core.archive_management._get_stdout',
                   return_value='testfile: (0x0000000d) released exists archived, archive_id:1'):
            assert archive_states('testfile') == ['released', 'exists', 'archived']

        with patch('egcg_core.archive_management._get_stdout',
                   return_value='testfile: (0x00000009) exists archived, archive_id:1'):
            assert archive_states('testfile') == ['exists', 'archived']

        with patch('egcg_core.archive_management._get_stdout',
                   return_value='testfile: (0x00000001) exists, archive_id:1'):
            assert archive_states('testfile') == ['exists']

        with patch('egcg_core.archive_management._get_stdout',
                   return_value='testfile: (0x00000000)'):
            assert archive_states('testfile') == []

    def test_release_file_from_lustre(self):
        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x00000009) exists archived, archive_id:1',
                       '',
                       'testfile: (0x0000000d) released exists archived, archive_id:1'
                   ]) as get_stdout:
            assert release_file_from_lustre('testfile')
            assert get_stdout.call_count == 3
            assert get_stdout.call_args_list[1][0] == ('lfs hsm_release testfile',)

        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x0000000d) released exists archived, archive_id:1',
                   ]) as get_stdout:
            assert release_file_from_lustre('testfile')
            assert get_stdout.call_count == 1

        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x00000009) exists, archive_id:1',
                   ]) as get_stdout:
            self.assertRaises(ArchivingError, release_file_from_lustre, 'testfile')
            assert get_stdout.call_count == 1

    def test_register_for_archiving(self):
        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x00000001)',
                       '',
                       'testfile: (0x00000009) exists, archive_id:1',
                    ]) as get_stdout:
            assert register_for_archiving('testfile')
            assert get_stdout.call_count == 3
            assert get_stdout.call_args_list[1][0] == ('lfs hsm_archive testfile',)

        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x00000001) exists, archive_id:1',
                   ]) as get_stdout:
            assert register_for_archiving('testfile')
            assert get_stdout.call_count == 1

        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x00000001)',
                       '',
                       'testfile: (0x00000001)',
                    ]) as get_stdout:
            self.assertRaises(ArchivingError, register_for_archiving, 'testfile')
            assert get_stdout.call_count == 3
            assert get_stdout.call_args_list[1][0] == ('lfs hsm_archive testfile',)

    def test_recall_from_tape(self):
        with patch('egcg_core.archive_management._get_stdout',
                   side_effect=[
                       'testfile: (0x0000000d) released exists archived, archive_id:1',
                       '',
                   ]) as get_stdout:
            assert recall_from_tape('testfile')
            assert get_stdout.call_count == 2
            assert get_stdout.call_args_list[1][0] == ('lfs hsm_restore testfile',)

    def test_archive_directory(self):
        with patch('egcg_core.archive_management.register_for_archiving') as register:
            assert archive_directory(os.path.join(self.assets_path, 'fastqs'))
            assert register.call_count == 6



from os import environ
from os.path import join
import pytest
from tests import TestEGCG
from egcg_core.config import Configuration, ConfigError


class TestConfiguration(TestEGCG):
    def setUp(self):
        self.cfg = Configuration(self.etc_config)
        self.cfg2 = Configuration()

    def test_delayed_load(self):
        cfg2 = Configuration()
        assert cfg2.get('ncbi_cache') is None
        cfg2.load_config_file(self.etc_config)
        assert cfg2.get('ncbi_cache') == ':memory:'

    def test_delayed_load_with_env(self):
        cfg2 = Configuration(env_var='TESTENV')
        environ['TESTENV'] = 'another_env'
        assert cfg2.get('ncbi_cache') is None
        cfg2.load_config_file(self.etc_config)
        assert cfg2.get('ncbi_cache') == 'path/to/ncbi.sqlite'

    def test_find_config_file(self):
        existing_cfg_file = self.etc_config
        non_existing_cfg_file = join(self.etc, 'non_existent.yaml')
        assert self.cfg._find_config_file((non_existing_cfg_file, existing_cfg_file)) == existing_cfg_file
        assert self.cfg._find_config_file((non_existing_cfg_file,)) is None

    def test_load_config_file(self):
        existing_cfg_file = self.etc_config
        non_existing_cfg_file = join(self.etc, 'non_existent.yaml')

        original_content = self.cfg.content
        self.cfg.content = {}
        self.cfg.load_config_file(existing_cfg_file)
        assert self.cfg.content == original_content

        self.cfg.content = {}
        with pytest.raises(ConfigError) as e:
            self.cfg.load_config_file(non_existing_cfg_file)
            assert str(e) == 'Could not find any config file in specified search path'

    def test_load_config_file_with_env(self):
        expected_content = self.cfg.content
        self.cfg.content = {}
        environ['TESTENV'] = 'another_env'
        self.cfg.load_config_file(self.etc_config, env_var='TESTENV')
        expected_content['ncbi_cache'] = 'path/to/ncbi.sqlite'
        expected_content.pop('another_env')
        assert self.cfg.content == expected_content

    def test_get(self):
        assert self.cfg.get('nonexistent_thing') is None
        assert self.cfg.get('nonexistent_thing', 'a_default') == 'a_default'
        assert self.cfg.get('executor').get('job_execution') == 'local'

    def test_query(self):
        assert self.cfg.query('executor', 'job_execution') == 'local'
        assert self.cfg.query('nonexistent_thing') is None
        assert self.cfg.query('logging', 'handlers', 'nonexistent_handler') is None
        assert self.cfg.query('logging', 'datefmt') == '%Y-%b-%d %H:%M:%S'

    def test_merge_dicts(self):
        default_dict = {
            'this': {'another': [2, 3, 4], 'more': {'thing': 'thang'}},
            'that': 'a_thing',
            'other': {'another': [2, '3', 4], 'more': {'thing': 'thang'}}
        }
        override_dict = {
            'that': 'another_thing',
            'another': 4,
            'more': 5,
            'other': {'another': [8, 9, 10], 'more': {'thung': 'theng'}}
        }
        merged_dict = self.cfg._merge_dicts(default_dict, override_dict)

        assert dict(merged_dict) == {
            'this': {'another': [2, 3, 4], 'more': {'thing': 'thang'}},
            'that': 'another_thing',
            'other': {'another': [8, 9, 10], 'more': {'thing': 'thang', 'thung': 'theng'}},
            'another': 4,
            'more': 5
        }

        assert dict(self.cfg._merge_dicts(default_dict, default_dict)) == default_dict

    def test_contains(self):
        assert 'rest_api' in self.cfg



import re
from genologics.lims import Lims
from egcg_core.config import cfg
from egcg_core.app_logging import logging_default as log_cfg
from egcg_core.exceptions import EGCGError

app_logger = log_cfg.get_logger('clarity')
try:
    from egcg_core.ncbi import get_species_name
except ImportError:
    app_logger.warning('Could not import egcg_core.ncbi. Is sqlite3 available?')

    def get_species_name(query_species):
        raise EGCGError('Could not import egcg_core.ncbi.get_species_name - sqlite3 seems to be unavailable.')


_lims = None


def connection():
    global _lims
    if not _lims:
        _lims = Lims(**cfg.get('clarity'))
    return _lims


def get_valid_lanes(flowcell_name):
    """
    Get all valid lanes for a given flowcell
    :param str flowcell_name: a flowcell id, e.g. HCH25CCXX
    :return: list of numbers of non-failed lanes
    """
    containers = connection().get_containers(type='Patterned Flowcell', name=flowcell_name)
    if len(containers) != 1:
        app_logger.warning('%s Flowcell(s) found for name %s', len(containers), flowcell_name)
        return None

    flowcell = containers[0]
    valid_lanes = []
    for placement_key in flowcell.placements:
        lane = int(placement_key.split(':')[0])
        artifact = flowcell.placements.get(placement_key)
        if not artifact.udf.get('Lane Failed?', False):
            valid_lanes.append(lane)
    valid_lanes = sorted(valid_lanes)
    app_logger.info('Valid lanes for %s: %s', flowcell_name, str(valid_lanes))
    return valid_lanes


def find_project_name_from_sample(sample_name):
    samples = get_samples(sample_name)
    if samples:
        project_names = set([s.project.name for s in samples])
        if len(project_names) == 1:
            return project_names.pop()
        else:
            app_logger.error('%s projects found for sample %s', len(project_names), sample_name)


def find_run_elements_from_sample(sample_name):
    sample = get_sample(sample_name)
    if sample:
        run_log_files = connection().get_artifacts(
            sample_name=sample.name,
            process_type='AUTOMATED - Sequence'
        )
        for run_log_file in run_log_files:
            p = run_log_file.parent_process
            run_id = p.udf.get('RunID')
            lanes = p.input_per_sample(sample.name)
            for artifact in lanes:
                lane = artifact.position.split(':')[0]
                if not artifact.udf.get('Lane Failed?', False):
                    yield run_id, lane


def get_species_from_sample(sample_name):
    samples = get_samples(sample_name)
    if samples:
        species_strings = set([s.udf.get('Species') for s in samples])
        nspecies = len(species_strings)
        if nspecies != 1:
            app_logger.error('%s species found for sample %s', nspecies, sample_name)
            return None
        species_string = species_strings.pop()
        if species_string:
            return get_species_name(species_string)


def get_genome_version(sample_id, species=None):
    s = get_sample(sample_id)
    if not s:
        return None
    genome_version = s.udf.get('Genome Version', None)
    if not genome_version and species:
        return cfg.query('species', species, 'default')
    return genome_version


def sanitize_user_id(user_id):
    if isinstance(user_id, str):
        return re.sub("[^\w_\-.]", "_", user_id)


substitutions = (
    (None, None),
    (re.compile('_(\d{2})$'), ':\g<1>'),  # '_01' -> ':01'
    (re.compile('__(\w):(\d{2})'), ' _\g<1>:\g<2>')  # '__L:01' -> ' _L:01'
)


def get_list_of_samples(sample_names):
    max_query = 100
    results = []
    for start in range(0, len(sample_names), max_query):
        results.extend(_get_list_of_samples(sample_names[start:start+max_query]))
    return results


def _get_list_of_samples(sample_names, sub=0):
    pattern, repl = substitutions[sub]
    _sample_names = list(sample_names)
    if pattern and repl:
        _sample_names = [pattern.sub(repl, s) for s in _sample_names]

    lims = connection()
    samples = lims.get_samples(name=_sample_names)
    lims.get_batch(samples)

    if len(samples) != len(sample_names):  # haven't got all the samples because some had _01/__L:01
        sub += 1
        remainder = sorted(set(_sample_names).difference(set([s.name for s in samples])))
        if sub < len(substitutions):
            samples.extend(_get_list_of_samples(remainder, sub))
        else:  # end recursion
            app_logger.warning('Could not find %s in Lims' % remainder)

    return samples


def get_samples(sample_name):
    lims = connection()
    samples = lims.get_samples(name=sample_name)
    # FIXME: Remove the hack when we're sure our sample id don't have colon
    if len(samples) == 0:
        sample_name_sub = re.sub("_(\d{2})$", ":\g<1>", sample_name)
        samples = lims.get_samples(name=sample_name_sub)
    if len(samples) == 0:
        sample_name_sub = re.sub("__(\w)_(\d{2})", " _\g<1>:\g<2>", sample_name)
        samples = lims.get_samples(name=sample_name_sub)
    return samples


def get_sample(sample_name):
    samples = get_samples(sample_name)
    if len(samples) != 1:
        app_logger.warning('%s Sample(s) found for name %s', len(samples), sample_name)
        return None
    return samples[0]


def get_user_sample_name(sample_name, lenient=False):
    """
    Query the LIMS and return the name the user gave to the sample
    :param str sample_name: our internal sample ID
    :param bool lenient: If True, return the sample name if no user sample name found
    :return: the user's sample name, None or the input sample name
    """
    user_sample_name = get_sample(sample_name).udf.get('User Sample Name')
    if user_sample_name:
        return sanitize_user_id(user_sample_name)
    elif lenient:
        return sample_name


def get_sample_gender(sample_name):
    sample = get_sample(sample_name)
    if sample:
        gender = sample.udf.get('Sex')
        if not gender:
            gender = sample.udf.get('Gender')
        return gender


def get_sample_genotype(sample_name, output_file_name):
    sample = get_sample(sample_name)
    if sample:
        file_id = sample.udf.get('Genotyping results file id')
        if file_id:
            file_content = connection().get_file_contents(id=file_id)
            with open(output_file_name, 'w') as open_file:
                open_file.write(file_content)
            return output_file_name
        else:
            app_logger.warning('Cannot download genotype results for %s', sample_name)


def get_expected_yield_for_sample(sample_name):
    """
    Query the LIMS and return the number of bases expected for a sample
    :param sample_name: the sample name
    :return: number of bases
    """
    sample = get_sample(sample_name)
    if sample:
        nb_gb = sample.udf.get('Yield for Quoted Coverage (Gb)')
        if nb_gb:
            return nb_gb * 1000000000


def get_run(run_id):
    runs = connection().get_processes(type='AUTOMATED - Sequence', udf={'RunID': run_id})
    if not runs:
        return None
    elif len(runs) != 1:
        app_logger.error('%s runs found for %s', len(runs), run_id)
    return runs[0]


def route_samples_to_delivery_workflow(sample_names):
    lims = connection()
    samples = [get_sample(sample_name) for sample_name in sample_names]
    artifacts = [sample.artifact for sample in samples]
    workflow_uri = lims.get_uri('configuration', 'workflows', '401')
    lims.route_artifacts(artifacts, workflow_uri=workflow_uri)


def get_plate_id_and_well(sample_name):
    sample = get_sample(sample_name)
    if sample:
        plate, well = sample.artifact.location
        return plate.name, well
    else:
        return None, None


def get_sample_names_from_plate(plate_id):
    containers = connection().get_containers(type='96 well plate', name=plate_id)
    if containers:
        samples = {}
        placements = containers[0].get_placements()
        for key in placements:
            sample_name = placements.get(key).samples[0].name
            samples[key] = sanitize_user_id(sample_name)
        return list(samples.values())


def get_sample_names_from_project(project_id):
    samples = connection().get_samples(projectname=project_id)
    sample_names = [sample.name for sample in samples]
    return sample_names


def get_output_containers_from_sample_and_step_name(sample_name, step_name):
    lims = connection()
    sample = get_sample(sample_name)
    sample_name = sample.name
    containers = set()
    arts = [a.id for a in lims.get_artifacts(sample_name=sample_name)]
    prcs = lims.get_processes(type=step_name, inputartifactlimsid=arts)
    for prc in prcs:
        arts = prc.input_per_sample(sample_name)
        for art in arts:
            containers.update([o.container for o in prc.outputs_per_input(art.id, Analyte=True)])
    return containers


def get_samples_arrived_with(sample_name):
    sample = get_sample(sample_name)
    samples = set()
    if sample:
        container = sample.artifact.container
        if container.type.name == '96 well plate':
            samples = get_sample_names_from_plate(container.name)
    return samples


def get_samples_for_same_step(sample_name, step_name):
    sample = get_sample(sample_name)
    sample_name = sample.name
    containers = get_output_containers_from_sample_and_step_name(sample_name, step_name)
    samples = set()
    for container in containers:
        samples.update(get_sample_names_from_plate(container.name))
    return samples


def get_samples_genotyped_with(sample_name):
    return get_samples_for_same_step(sample_name, 'Genotyping Plate Preparation EG 1.0')


def get_samples_sequenced_with(sample_name):
    return get_samples_for_same_step(sample_name, 'Sequencing Plate Preparation EG 1.0')


def get_released_samples():
    released_samples = []
    processes = connection().get_processes(type='Data Release EG 1.0')
    for process in processes:
        for artifact in process.all_inputs():
            released_samples.extend([sanitize_user_id(s.name) for s in artifact.samples])

    return sorted(set(released_samples))


def get_sample_release_date(sample_id):
    s = get_sample(sample_id)
    if not s:
        return None
    procs = connection().get_processes(type='Data Release EG 1.0', inputartifactlimsid=s.artifact.id)
    if not procs:
        return None
    elif len(procs) != 1:
        app_logger.warning('%s Processes found for sample %s: Return latest one', len(procs), sample_id)
        return sorted([p.date_run for p in procs], reverse=True)[0]
    return procs[0].date_run


def get_project(project_id):
    lims = connection()
    projects = lims.get_projects(name=project_id)
    if len(projects) != 1:
        app_logger.warning('%s Project(s) found for name %s', len(projects), project_id)
        return None
    return projects[0]



import jinja2
import smtplib
from time import sleep
from email.mime.text import MIMEText
from egcg_core.exceptions import EGCGError
from .notification import Notification


class EmailNotification(Notification):
    translation_map = {' ': '&nbsp', '\n': '<br/>'}

    def __init__(self, name, mailhost, port, sender, recipients, strict=False, email_template=None):
        super().__init__(name)
        self.mailhost = mailhost
        self.port = port
        self.sender = sender
        self.recipients = recipients
        self.email_template = email_template
        self.strict = strict

    def notify(self, msg):
        email = self.build_email(msg)
        success = self._try_send(email)
        if not success:
            err_msg = 'Failed to send message: ' + str(msg)
            if self.strict:
                raise EGCGError(err_msg)
            else:
                self.critical(err_msg)

    def _try_send(self, msg, retries=3):
        """
        Prepare a MIMEText message from body and diagnostics, and try to send a set number of times.
        :param int retries: Which retry we're currently on
        :return: True if a message is sucessfully sent, otherwise False
        """
        try:
            self._connect_and_send(msg)
            return True
        except (smtplib.SMTPException, TimeoutError) as e:
            retries -= 1
            self.warning('Encountered a %s exception. %s retries remaining', str(e), retries)
            if retries:
                sleep(2)
                return self._try_send(msg, retries)

            return False

    def build_email(self, body):
        """
        Use Jinja to build a MIMEText html-formatted email from plain text.
        :param str body: The main body of the email to send
        """
        if self.email_template:
            content = jinja2.Template(open(self.email_template).read())
            msg = MIMEText(content.render(title=self.name, body=self._prepare_string(body)), 'html')
        else:
            msg = MIMEText(body)

        msg['Subject'] = self.name
        msg['From'] = self.sender
        msg['To'] = ', '.join(self.recipients)
        return msg

    @classmethod
    def _prepare_string(cls, msg):
        for k in cls.translation_map:
            msg = msg.replace(k, cls.translation_map[k])
        return msg

    def _connect_and_send(self, msg):
        connection = smtplib.SMTP(self.mailhost, self.port)
        connection.send_message(msg, self.sender, self.recipients)
        connection.quit()


def send_email(msg, mailhost, port, sender, recipients, subject, email_template=None, strict=False):
    EmailNotification(subject, mailhost, port, sender, recipients, email_template, strict).notify(msg)




ELEMENT_RUN_ELEMENT_ID = 'run_element_id'
ELEMENT_RUN_NAME = 'run_id'
ELEMENT_LANE_ID = 'lane_id'
ELEMENT_LANE = 'lane'
ELEMENT_LANE_NUMBER = 'lane_number'
ELEMENT_NUMBER_LANE = 'number_of_lanes'
ELEMENT_BARCODE = 'barcode'
ELEMENT_PROJECT = 'project'
ELEMENT_PROJECT_ID = 'project_id'
ELEMENT_LIBRARY_INTERNAL_ID = 'library_id'
ELEMENT_SAMPLE_INTERNAL_ID = 'sample_id'
ELEMENT_SAMPLES = 'samples'
ELEMENT_NB_READS_SEQUENCED = 'total_reads'
ELEMENT_NB_READS_PASS_FILTER = 'passing_filter_reads'
ELEMENT_PC_READ_IN_LANE = 'pc_reads_in_lane'
ELEMENT_NB_BASE_R1 = 'bases_r1'
ELEMENT_NB_BASE_R2 = 'bases_r2'
ELEMENT_NB_Q30_R1 = 'q30_bases_r1'
ELEMENT_NB_Q30_R2 = 'q30_bases_r2'
ELEMENT_NB_READS_CLEANED = 'clean_reads'  # this implies that the filter has been passed
ELEMENT_NB_BASE_R1_CLEANED = 'clean_bases_r1'
ELEMENT_NB_BASE_R2_CLEANED = 'clean_bases_r2'
ELEMENT_NB_Q30_R1_CLEANED = 'clean_q30_bases_r1'
ELEMENT_NB_Q30_R2_CLEANED = 'clean_q30_bases_r2'
ELEMENT_LANE_PC_OPT_DUP = 'lane_pc_optical_dups'
ELEMENT_FASTQC_REPORT_R1 = 'fastqc_report_r1'
ELEMENT_FASTQC_REPORT_R2 = 'fastqc_report_r2'
ELEMENT_ADAPTER_TRIM_R1 = 'adaptor_bases_removed_r1'
ELEMENT_ADAPTER_TRIM_R2 = 'adaptor_bases_removed_r2'

ELEMENT_SAMPLE_EXTERNAL_ID = 'user_sample_id'
ELEMENT_SAMPLE_SPECIES = 'species_name'
ELEMENT_SAMPLE_PLATE = 'plate_name'
ELEMENT_SAMPLE_EXPECTED_COVERAGE = 'expected_coverage'
ELEMENT_SAMPLE_EXPECTED_YIELD = 'expected_yield'
ELEMENT_SAMPLE_GENOME_SIZE = 'genome_size'
ELEMENT_NB_READS_IN_BAM = 'bam_file_reads'
ELEMENT_NB_MAPPED_READS = 'mapped_reads'
ELEMENT_NB_SEC_MAPPED_READS = 'Nb secondary alignments'
ELEMENT_NB_DUPLICATE_READS = 'duplicate_reads'
ELEMENT_NB_PROPERLY_MAPPED = 'properly_mapped_reads'
ELEMENT_MEDIAN_INSERT_SIZE = 'Median Insert Size'
ELEMENT_MEAN_INSERT_SIZE = 'Mean Insert Size'
ELEMENT_MEDIAN_COVERAGE = 'median_coverage'
ELEMENT_SAMTOOLS_MEDIAN_COVERAGE = 'samtools_median_coverage'
ELEMENT_PC_BASES_CALLABLE = 'pc_callable'
ELEMENT_RUN_ELEMENTS = 'run_elements'
ELEMENT_GENDER_VALIDATION = 'gender_validation'
ELEMENT_GENDER_HETX = 'hetX'
ELEMENT_GENDER_COVY = 'covY'
ELEMENT_CALLED_GENDER = 'called_gender'
ELEMENT_PROVIDED_GENDER = 'provided_gender'
ELEMENT_GENOTYPE_VALIDATION = 'genotype_validation'
ELEMENT_NO_CALL_CHIP = 'no_call_chip'
ELEMENT_NO_CALL_SEQ = 'no_call_seq'
ELEMENT_MATCHING = 'matching_snps'
ELEMENT_MISMATCHING = 'mismatching_snps'
ELEMENT_SPECIES_CONTAMINATION = 'species_contamination'
ELEMENT_SAMPLE_CONTAMINATION = 'sample_contamination'
ELEMENT_FREEMIX = 'freemix'
ELEMENT_SNPS_TI_TV = 'ti_tv_ratio'
ELEMENT_SNPS_HET_HOM = 'het_hom_ratio'
ELEMENT_TOTAL_READS_MAPPED = 'total_reads_mapped'
ELEMENT_CONTAMINANT_UNIQUE_MAP = 'contaminant_unique_mapped'
ELEMENT_PCNT_UNMAPPED_FOCAL = 'percent_unmapped_focal'
ELEMENT_PCNT_UNMAPPED = 'percent_unmapped'
ELEMENT_COVERAGE_STATISTICS = 'coverage'
ELEMENT_MEAN_COVERAGE = 'mean'
ELEMENT_MEDIAN_COVERAGE_SAMTOOLS = 'median'
ELEMENT_COVERAGE_PERCENTILES = 'coverage_percentiles'
ELEMENT_BASES_AT_COVERAGE = 'bases_at_coverage'
ELEMENT_PERCENTILE_5 = 'percentile_5'
ELEMENT_PERCENTILE_25 = 'percentile_25'
ELEMENT_PERCENTILE_50 = 'percentile_50'
ELEMENT_PERCENTILE_75 = 'percentile_75'
ELEMENT_PERCENTILE_95 = 'percentile_95'
ELEMENT_BASES_AT_5X = 'bases_at_5X'
ELEMENT_BASES_AT_15X = 'bases_at_15X'
ELEMENT_BASES_AT_30X = 'bases_at_30X'
ELEMENT_COVERAGE_SD = 'std_dev'
ELEMENT_COVERAGE_EVENNESS = 'evenness'
ELEMENT_REVIEWED = 'reviewed'
ELEMENT_REVIEW_COMMENTS = 'review_comments'
ELEMENT_REVIEW_DATE = 'review_date'
ELEMENT_USEABLE = 'useable'
ELEMENT_USEABLE_COMMENTS = 'useable_comments'
ELEMENT_USEABLE_DATE = 'useable_date'
ELEMENT_DELIVERED = 'delivered'
ELEMENT_DELIVERY_DATE = 'delivery_date'

ELEMENT_FASTQS_DELETED = 'input_fastqs_deleted'

ELEMENT_PROCS = 'analysis_driver_procs'
ELEMENT_PROC_ID = 'proc_id'
ELEMENT_STATUS = 'status'

DATASET_NEW = 'new'
DATASET_READY = 'ready'
DATASET_FORCE_READY = 'force_ready'
DATASET_RESUME = 'resume'
DATASET_PROCESSING = 'processing'
DATASET_PROCESSED_SUCCESS = 'finished'
DATASET_PROCESSED_FAIL = 'failed'
DATASET_ABORTED = 'aborted'
DATASET_REPROCESS = 'reprocess'
DATASET_DELETED = 'deleted'



from os import getenv
from os.path import isfile
import yaml
from .exceptions import ConfigError


class Configuration:
    config_file = None
    content = {}

    def __init__(self, *search_path, env_var=None):
        self.env_var = env_var
        self.config_file = None
        self.content = {}
        if search_path:
            self.load_config_file(*search_path, env_var=env_var)
            self._select_env()

    @staticmethod
    def _find_config_file(search_path):
        for p in search_path:
            if p and isfile(p):
                return p

    def load_config_file(self, *search_path, env_var=None):
        if env_var:
            self.env_var = env_var
        self.config_file = self._find_config_file(search_path)
        if self.config_file:
            self.content = yaml.safe_load(open(self.config_file, 'r'))
            self._select_env()
        else:
            raise ConfigError('Could not find any config file in specified search path')

    def get(self, item, ret_default=None):
        """
        Dict-style item retrieval with default
        :param item: The key to search for
        :param ret_default: What to return if the key is not present
        """
        try:
            return self[item]
        except KeyError:
            return ret_default

    def query(self, *parts, ret_default=None):
        """
        Drill down into a config, e.g. cfg.query('logging', 'handlers', 'a_handler', 'level')
        :param ret_default:
        :return: The relevant item if it exists in the config, else ret_default.
        """
        top_level = self.content
        item = None

        for p in parts:
            item = top_level.get(p)
            if item:
                top_level = item
            else:
                return ret_default

        return item

    def report(self):
        return yaml.safe_dump(self.content, default_flow_style=False)

    def __getitem__(self, item):
        """Allow dict-style access, e.g. config['this'] or config['this']['that']."""
        return self.content[item]

    def __contains__(self, item):
        """Allow search in the first layer of the config with 'in' operator."""
        return self.content.__contains__(item)

    def _select_env(self):
        # for backward compatibility keep merging the default section in the top level
        if 'default' in self.content:
            self.content = dict(self._merge_dicts(self.content, self.content['default']))
            self.content.pop('default')

        if self.env_var:
            env = getenv(self.env_var)
            if env and env in self.content:
                self.content = dict(self._merge_dicts(self.content, self.content[env]))
                self.content.pop(env)

    @classmethod
    def _merge_dicts(cls, default_dict, override_dict):
        """Recursively merge a default dict and an overriding dict."""
        for k in set(override_dict.keys()).union(default_dict.keys()):
            if k in default_dict and k in override_dict:
                if type(default_dict[k]) is dict and type(override_dict[k]) is dict:
                    yield k, dict(cls._merge_dicts(default_dict[k], override_dict[k]))
                else:
                    yield k, override_dict[k]
            elif k in default_dict:
                yield k, default_dict[k]
            else:
                yield k, override_dict[k]

    def merge(self, override_dict):
        """
        Merge the provided dict with the config content, potentially overriding existing parameters
        :param dict override_dict:
        """
        self.content = dict(self._merge_dicts(self.content, override_dict))


# For backward compatibility
EnvConfiguration = Configuration

cfg = Configuration()



import hashlib
import os
from os import makedirs
from shutil import rmtree
from os.path import join, basename
from tests import TestEGCG
from egcg_core import util

fastq_dir = join(TestEGCG.assets_path, 'fastqs')


def test_find_files():
    expected = [join(TestEGCG.assets_path, f) for f in ('ftest.txt', 'ftest_2.txt')]
    assert util.find_files(TestEGCG.assets_path, 'ftest*.txt') == expected


def test_find_file():
    assert util.find_file(TestEGCG.assets_path, 'ftest.txt') == join(TestEGCG.assets_path, 'ftest.txt')


def test_str_join():
    assert util.str_join('this', 'that', 'other', separator='/') == 'this/that/other'


def test_find_fastqs():
    fastqs = util.find_fastqs(fastq_dir, '10015AT', '10015AT0001')
    for file_name in ['10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz',
                      '10015AT0001_S6_L005_R1_001.fastq.gz', '10015AT0001_S6_L005_R2_001.fastq.gz']:
        assert join(fastq_dir, '10015AT', '10015AT0001', file_name) in fastqs


def test_find_fastqs_with_lane():
    fastqs = util.find_fastqs(fastq_dir, '10015AT', '10015AT0001', lane=4)
    for file_name in ['10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz']:
        assert join(fastq_dir, '10015AT', '10015AT0001', file_name) in fastqs


def test_find_all_fastqs():
    fastqs = util.find_all_fastqs(fastq_dir)
    for file_name in ('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz',
                      '10015AT0002_merged_R1.fastq.gz', '10015AT0002_merged_R2.fastq.gz'):
        assert file_name in [basename(f) for f in fastqs]


def test_find_all_fastq_pairs():
    observed = util.find_all_fastq_pairs(join(fastq_dir, '10015AT', '10015AT0001'))
    expected = [('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz'),
                ('10015AT0001_S6_L005_R1_001.fastq.gz', '10015AT0001_S6_L005_R2_001.fastq.gz')]
    assert [(basename(f), basename(g)) for f, g in observed] == expected


def test_same_fs():
    test = join(TestEGCG.assets_path, 'ftest.txt')
    test_2 = join(TestEGCG.assets_path, 'ftest_2.txt')
    test_nonexistent = join(TestEGCG.assets_path, 'ftest_nonexistent.txt')

    assert util.same_fs(test, None) is False
    assert util.same_fs(test, test_2)
    assert util.same_fs(test, test_nonexistent)


class TestMoveDir(TestEGCG):

    def _create_test_file(self, f, content=None):
        with open(f, 'w') as of:
            if content:
                of.write(content)
            else:
                of.write('This is a test file')

    def _md5(self, f):
        hash_md5 = hashlib.md5()
        with open(f, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def setUp(self):
        self.test_dir = join(self.assets_path, 'move_dir')
        makedirs(join(self.test_dir, 'from'), exist_ok=True)
        makedirs(join(self.test_dir, 'from', 'subdir'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'from', 'ftest.txt'))
        self._create_test_file(join(self.test_dir, 'from', 'subdir', 'ftest.txt'))

        makedirs(join(self.test_dir, 'external'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'external', 'external.txt'), 'External file')
        os.symlink(join(self.test_dir, 'external', 'external.txt'), join(self.test_dir, 'from', 'external_renamed.txt'))

        makedirs(join(self.test_dir, 'exists'), exist_ok=True)
        makedirs(join(self.test_dir, 'exists', 'subdir'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'exists', 'subdir', 'ftest.txt'), 'another file')
        self._create_test_file(join(self.test_dir, 'exists', 'ftest.txt'), 'another file')

    def tearDown(self):
        rmtree(util.find_file(self.test_dir, 'to'))
        rmtree(util.find_file(self.test_dir, 'from'))
        rmtree(util.find_file(self.test_dir, 'exists'))
        rmtree(util.find_file(self.test_dir, 'external'))

    def test_move_dir(self):
        frm = join(self.test_dir, 'from')
        to = join(self.test_dir, 'to')
        md5_from = self._md5(join(frm, 'ftest.txt'))
        assert util.find_file(frm, 'ftest.txt')
        assert not util.find_file(to)

        assert util.move_dir(frm, to) == 0

        assert not util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert util.find_file(to, 'subdir', 'ftest.txt')
        assert md5_from == self._md5(join(to, 'ftest.txt'))

        assert util.find_file(to, 'external_renamed.txt')

    def _move_dir_exists(self):
        frm = join(self.test_dir, 'from')
        to = join(self.test_dir, 'exists')
        md5_from1 = self._md5(join(frm, 'ftest.txt'))
        md5_from2 = self._md5(join(frm, 'subdir', 'ftest.txt'))

        assert util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert not md5_from1 == self._md5(join(to, 'ftest.txt'))
        assert not md5_from2 == self._md5(join(to, 'subdir', 'ftest.txt'))

        util.move_dir(frm, to)

        assert not util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert md5_from1 == self._md5(join(to, 'ftest.txt'))
        assert md5_from2 == self._md5(join(to, 'subdir', 'ftest.txt'))



from os.path import join, abspath, dirname
import pytest
from unittest.mock import Mock, patch
from smtplib import SMTPException
from email.mime.text import MIMEText
from egcg_core.notifications import NotificationCentre, EmailNotification, send_email, AsanaNotification
from egcg_core.exceptions import EGCGError
from tests import TestEGCG


class FakeSMTP(Mock):
    def __init__(self, host, port):
        super().__init__()
        self.mailhost, self.port = host, port

    @staticmethod
    def send_message(msg, reporter, recipients):
        if 'dodgy' in str(msg):
            raise SMTPException('Oh noes!')
        else:
            pass

    @staticmethod
    def quit():
        pass


class TestNotificationCentre(TestEGCG):
    def setUp(self):
        self.notification_centre = NotificationCentre('a_name')

    def test_config_init(self):
        e = self.notification_centre.subscribers['email']
        assert e.sender == 'this'
        assert e.recipients == ['that', 'other']
        assert e.mailhost == 'localhost'
        assert e.port == 1337
        assert e.strict is True

        a = self.notification_centre.subscribers['asana']
        assert a.workspace_id == 1337
        assert a.project_id == 1338

    def test_notify(self):
        self.notification_centre.subscribers = {'asana': Mock(), 'email': Mock()}
        self.notification_centre.notify_all('a message')
        for name, s in self.notification_centre.subscribers.items():
            s.notify.assert_called_with('a message')


class TestEmailNotification(TestEGCG):
    def setUp(self):
        self.ntf = EmailNotification(
            'a_subject', 'localhost', 1337, 'a_sender', ['some', 'recipients'], strict=True,
            email_template=join(dirname(dirname(abspath(__file__))), 'etc', 'email_notification.html')
        )

    @patch('egcg_core.notifications.EmailNotification._logger')
    def test_retries(self, mocked_logger):
        with patch('smtplib.SMTP', new=FakeSMTP), patch('egcg_core.notifications.email.sleep'):
            assert self.ntf._try_send(self.ntf.build_email('this is a test')) is True
            assert self.ntf._try_send(self.ntf.build_email('dodgy')) is False
            for i in range(3):
                mocked_logger.warning.assert_any_call(
                    'Encountered a %s exception. %s retries remaining', 'Oh noes!', i
                )

            with pytest.raises(EGCGError) as e:
                self.ntf.notify('dodgy')
                assert 'Failed to send message: dodgy' in str(e)

    def test_build_email(self):
        exp_msg = (
            '<!DOCTYPE html>\n'
            '<html lang="en">\n'
            '<head>\n'
            '    <meta charset="UTF-8">\n'
            '    <style>\n'
            '        table, th, td {\n'
            '            border: 1px solid black;\n'
            '            border-collapse: collapse;\n'
            '        }\n'
            '    </style>\n'
            '</head>\n'
            '<body>\n'
            '    <h2>a_subject</h2>\n'
            '    <p>a&nbspmessage</p>\n'
            '</body>\n'
            '</html>'
        )

        exp = MIMEText(exp_msg, 'html')
        exp['Subject'] = 'a_subject'
        exp['From'] = 'a_sender'
        exp['To'] = 'some, recipients'
        obs = self.ntf.build_email('a message')
        assert str(obs) == str(exp)

    def test_build_email_plain_text(self):
        self.ntf.email_template = None
        exp = MIMEText('a message')
        exp['Subject'] = 'a_subject'
        exp['From'] = 'a_sender'
        exp['To'] = 'some, recipients'
        obs = self.ntf.build_email('a message')
        assert str(obs) == str(exp)


@patch('egcg_core.notifications.EmailNotification._try_send')
def test_send_email(mocked_send):
    send_email('a message', 'localhost', 1337, 'a_sender', ['some', 'recipients'], 'a subject')
    exp = MIMEText('a message')
    exp['Subject'] = 'a subject'
    exp['From'] = 'a_sender'
    exp['To'] = 'some, recipients'
    assert str(mocked_send.call_args[0][0]) == str(exp)


class TestAsanaNotification(TestEGCG):
    def setUp(self):
        self.ntf = AsanaNotification(
            'another_name',
            workspace_id=1337,
            project_id=1338,
            access_token='an_access_token'
        )
        self.ntf.client = Mock(
            tasks=Mock(
                find_all=Mock(return_value=[{'name': 'this'}]),
                create_in_workspace=Mock(return_value={'id': 1337}),
                find_by_id=Mock(return_value={'name': 'this', 'id': 1337})
            )
        )

    def test_task(self):
        assert self.ntf.task == {'id': 1337, 'name': 'this'}
        self.ntf.client.tasks.find_by_id.assert_called_with(1337)

    def test_add_comment(self):
        self.ntf.notify('a comment')
        self.ntf.client.tasks.add_comment.assert_called_with(1337, text='a comment')

    def test_get_entity(self):
        collection = [{'name': 'this'}, {'name': 'that'}]
        assert self.ntf._get_entity(collection, 'that') == {'name': 'that'}
        assert self.ntf._get_entity(collection, 'other') is None

    def test_create_task(self):
        assert self.ntf._create_task() == {'id': 1337}
        self.ntf.client.tasks.create_in_workspace.assert_called_with(1337, self.ntf.task_template)



import os
import pytest
import shutil
import subprocess
from unittest.mock import patch, Mock
from tests import TestEGCG
from egcg_core.executor import Executor, StreamExecutor, ArrayExecutor, PBSExecutor, SlurmExecutor
from egcg_core.executor.cluster_executor import ClusterExecutor, running_executors, stop_running_jobs
from egcg_core.exceptions import EGCGError

get_stdout = 'egcg_core.executor.cluster_executor.ClusterExecutor._get_stdout'
sleep = 'egcg_core.executor.cluster_executor.sleep'


class TestExecutor(TestEGCG):
    def test_cmd(self):
        e = Executor('ls ' + os.path.join(self.assets_path, '..'))
        exit_status = e.join()
        assert exit_status == 0

    def test_dodgy_cmd(self):
        with pytest.raises(EGCGError) as err:
            e = Executor('dodgy_cmd')
            e.join()
            assert 'Command failed: \'dodgy_cmd\'' in str(err)

    def test_process(self):
        e = Executor('ls ' + os.path.join(self.assets_path, '..'))
        assert e.proc is None
        proc = e._process()
        assert proc is e.proc and isinstance(e.proc, subprocess.Popen)


class TestStreamExecutor(TestExecutor):
    def test_cmd(self):
        e = StreamExecutor(os.path.join(self.assets_path, 'countdown.sh'))
        e.start()
        assert e.join() == 0

    def test_dodgy_command(self):
        e = StreamExecutor(os.path.join(self.assets_path, 'countdown.sh') + ' dodgy')
        e.start()
        assert e.join() == 13  # same exit status as the running script

    def test_dodgy_cmd(self):
        with pytest.raises(EGCGError) as err:
            e = StreamExecutor('dodgy_cmd')
            e.start()
            e.error = Mock()
            e.join()
            assert 'self.proc command failed: \'dodgy_cmd\'' in str(err)


class TestArrayExecutor(TestExecutor):
    def test_cmd(self):
        e = ArrayExecutor(['ls', 'ls -lh', 'pwd'], stream=True)
        e.start()
        assert e.join() == 0
        assert e.exit_statuses == [0, 0, 0]

    def test_dodgy_cmd(self):
        e = ArrayExecutor(['ls', 'non_existent_cmd', 'pwd'], stream=True)
        for s in e.executors:
            s.error = Mock()

        e.error = Mock()
        e.start()
        with pytest.raises(EGCGError) as err:
            e.join()

        assert 'Commands failed' in str(err)
        e.error.assert_called_with('EGCGError: self.proc command failed: non_existent_cmd')


class TestClusterExecutor(TestEGCG):
    ppath = 'egcg_core.executor.cluster_executor.ClusterExecutor'
    script = os.path.join(TestEGCG.assets_path, 'countdown.sh')

    def setUp(self):
        os.makedirs(os.path.join(self.assets_path, 'a_run_id'), exist_ok=True)
        self.executor = ClusterExecutor(
            self.script,
            job_name='test_job',
            working_dir=os.path.join(self.assets_path, 'a_run_id')
        )

    def tearDown(self):
        shutil.rmtree(os.path.join(self.assets_path, 'a_run_id'))

    def test_get_stdout(self):
        popen = 'egcg_core.executor.executor.subprocess.Popen'
        with patch(popen, return_value=Mock(wait=Mock(return_value=None))) as p:
            assert self.executor._get_stdout('ls -d ' + self.assets_path).endswith('tests/assets')
            p.assert_called_with(['ls', '-d', self.assets_path], stdout=-1, stderr=-1)

    def test_run_and_retry(self):
        with patch(get_stdout, side_effect=[None, None, self.assets_path]) as p, patch(sleep):
            assert self.executor._run_and_retry('ls -d ' + self.assets_path).endswith('tests/assets')
            assert p.call_count == 3

    def test_dodgy_cmd(self):
        with pytest.raises(EGCGError) as err, patch(get_stdout, return_value=None), patch(sleep):
            self.executor.cmds = [os.path.join(self.assets_path, 'non_existent_script.sh')]
            self.executor.start()
            assert str(err) == 'Job submission failed'

    def test_join(self):
        job_finished = self.ppath + '._job_finished'
        exit_code = self.ppath + '._job_exit_code'
        self.executor.finished_statuses = 'FXM'
        with patch(job_finished, return_value=True), patch(exit_code, return_value=0), patch(sleep):
            assert self.executor.join() == 0

    def test_job_cancellation(self):
        with patch(self.ppath + '._submit_job'), patch(self.ppath + '._job_finished', return_value=True),\
             patch(self.ppath + '.write_script'), patch(self.ppath + '._job_exit_code', return_value=9),\
             patch(self.ppath + '.cancel_job'), patch(sleep):

            self.executor.job_id = 'test_job'
            self.executor.start()
            assert running_executors == {'test_job': self.executor}
            stop_running_jobs()
            assert running_executors == {}


class TestPBSExecutor(TestClusterExecutor):
    ppath = 'egcg_core.executor.cluster_executor.PBSExecutor'

    def setUp(self):
        os.makedirs(os.path.join(self.assets_path, 'a_run_id'), exist_ok=True)
        self.executor = PBSExecutor(
            self.script,
            job_name='test_job',
            working_dir=os.path.join(self.assets_path, 'a_run_id')
        )

    def test_qstat(self):
        fake_report = ('Job id            Name             User              Time Use S Queue\n'
                       '----------------  ---------------- ----------------  -------- - -------\n'
                       '1337[].server     a_job_name       a_user            0        B a_queue\n'
                       '1338.server       another_job_name another_user      00:00:00 R a_queue\n')
        with patch(get_stdout, return_value=fake_report) as p:
            assert self.executor._qstat() == [
                '1337[].server     a_job_name       a_user            0        B a_queue',
                '1338.server       another_job_name another_user      00:00:00 R a_queue'
            ]
            p.assert_called_with('qstat -xt None')

    def test_job_status(self):
        qstat = 'egcg_core.executor.cluster_executor.PBSExecutor._qstat'
        fake_report = ['1337.server   a_job   a_user   10:00:00   R    q']
        with patch(qstat, return_value=fake_report):
            assert self.executor._job_statuses() == {'R'}

    def test_job_finished(self):
        job_statuses = 'egcg_core.executor.cluster_executor.PBSExecutor._job_statuses'
        with patch(job_statuses, return_value={'F', 'M', 'X', 'B'}):
            assert not self.executor._job_finished()
        with patch(job_statuses, return_value={'F', 'F', 'M', 'X'}):
            assert self.executor._job_finished()


class TestSlurmExecutor(TestClusterExecutor):
    ppath = 'egcg_core.executor.cluster_executor.SlurmExecutor'

    def setUp(self):
        os.makedirs(os.path.join(self.assets_path, 'a_run_id'), exist_ok=True)
        self.executor = SlurmExecutor(
            self.script,
            job_name='test_job',
            working_dir=os.path.join(self.assets_path, 'a_run_id')
        )

    def test_sacct(self):
        with patch(get_stdout, return_value=' COMPLETED  0:0 \n COMPLETED  0:0\n FAILED 1:0') as p:
            assert self.executor._sacct('State,ExitCode') == {'COMPLETED  0:0', 'FAILED 1:0'}
            p.assert_called_with('sacct -nX -j None -o State,ExitCode')

    def test_squeue(self):
        with patch(get_stdout, return_value='RUNNING\nRUNNING\nRUNNING') as p:
            assert self.executor._squeue() == {'RUNNING'}
            p.assert_called_with('squeue -h -j None -o %T')

    def test_job_finished(self):
        sacct = 'egcg_core.executor.cluster_executor.SlurmExecutor._sacct'
        patched_squeue = patch('egcg_core.executor.cluster_executor.SlurmExecutor._squeue', return_value='')
        with patch(sacct, return_value=['RUNNING']), patched_squeue:
            assert not self.executor._job_finished()
        with patch(sacct, return_value=['COMPLETED']), patched_squeue:
            assert self.executor._job_finished()

    def test_job_exit_code(self):
        sacct = 'egcg_core.executor.cluster_executor.SlurmExecutor._sacct'
        with patch(sacct, return_value=['CANCELLED 0:0']):
            assert self.executor._job_exit_code() == 9
        with patch(sacct, return_value=['COMPLETED 0:x']):
            assert self.executor._job_exit_code() == 0
